
function [C,U,W]=BKSym(A,U0,W0,p,stage)
% Symmetric BK
% A is a 3-tensor, (1,2)-symmetric
% U0, U0, W0 have second dimension r1, r1, r3
% p : Block size
% stage : number of stages
% Output : 
% C : Core tensor 
% U=[U1 U2 ...] (as many as are generated in the stages executed 
% Same for  W 
% Check if U or W become too large in the second and third stage. If so avoid tensor-matrix multiplication.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Distributed under a 3-Clause BSD License 
%
% Copyright 2021 Lars Eldén, Linköping University, and Maryam Dehghan,
% Federal University of Ceará,
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

r1=size(U0,2); r3=size(W0,2);
n=size(A);

if stage>3, disp('Error, stage 4 not implemented yet'), stage=3; end
%%%%%%%%%%%%% stage 1 %%%%%%%%%%%%%%%%%

% U1 <-- U0,W0
[U,U1]=newblock(A,U0,U0,W0,[1 3],2);         %  [2 3],1 was much slower

% W1 <-- U0,U0                     % Due to symmetry, we should remove
                                   % about half the vectors in A.(U0,U0) before
                                   % computing W1.  
[W,W1]=triblockSym(A,W0,U0,U0,[1 2],3); 

% 
if stage>1
    %%%%%%%%%%%%%% stage 2 %%%%%%%%%%%%%%%%
    
    if size(U,2)<n(1)
        %U2 <-- U0,W1
        [U,U2]=newblock(A,U,U0,barX(W1,p),[1 3],2);     %  [2 3],1 was much slower
        
        if size(U,2)<n(1)
            %U3 <-- U1,W0
            [U,U3]=newblock(A,U,barX(U1,p),W0,[1 3],2);  % [2 3],1 was much slower
        end
    end
    
    if size(W,2)<n(3)
        %W2 <-- U0,U1
        [W,W2]=newblock(A,W,U0,barX(U1,p),[1 2],3);
    end
    
    if stage==3
        %%%%%%%%%%%%%%% Stage 3 simplified %%%%%%%%%%%%%%%
        if size(U,2)< n(1)
            % U4 <-- U0,W2
            if exist('W2')
                [U,U4]=newblock(A,U,U0,barX(W2,p),[1 3],2);
            end
            if size(U,2)< n(1)
                % U5 <-- U2,W0
                [U,U5]=newblock(A,U,barX(U2,p),W0,[1 3],2);
            end
        end
        if size(W,2)<n(3)
            % W3 <-- U0,U2
            [W,W3]=newblock(A,W,U0,barX(U2,p),[1 2],3);
            if size(W,2)<n(3)
                % W4 <-- U1,U2
                [W,W4]=newblock(A,W,barX(U1,p),barX(U2,p),[1 2],3);
            end
        end
    end
end

C=(ttm(A,{U,U,W},[1,2,3],'t'));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Xbar=barX(X,p)
if size(X,2)<p
    Xbar=X;
else
    Xbar=X(:,1:p);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Xall,Xnew]=newblock(A,X,Y,Z,modeYZ,modeX)
Xhat=tam(double(ttm(A,{Y,Z},modeYZ,'t')),modeX);
%[Xhat,rhat,perm]=qr(Xhat,0);  % Permute columns to make the first columns of Xnew
                              % the most significant of the new.
                              % Initial tests show that this does not have
                              % any noticeable impact on the convergence rate.
k=size(X,2);
n=size(A);
[Xall,rall]=qr([X Xhat],0);   % No pivoting here
if size(Xall,2)>n(modeX)      % Check if Xall has full modeX dimension
    Xall=Xall(:,1:n(modeX));
end
Xnew=Xall(:,k+1:end);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Xall,Xnew]=triblockSym(A,X,Y,Z,modeYZ,modeX)
% Y=Z is assumed => ttm(A,{Y,Z},[1 2],'t') is (1,2)-symmetric. Remove the
% extra vectors
n=size(A); 
X00=double(ttm(A,{Y,Z},modeYZ,'t'));   % X0 is r1 x r1 x size(A,3)
% Xhat=tam(double(ttm(A,{Y,Z},modeYZ,'t')),modeX);
%[Xhat,rhat,perm]=qr(Xhat,0);  % Permute columns to make the first columns of Xnew
                              % the most significant of the new.
                              % Initial tests show that this does not have
                              % any noticeable impact on the convergence rate.
r1=size(X00,1);                               
jj=0; 
for i=1:r1
    for j=i:r1
        jj=jj+1;
        Xhat(:,jj)=X00(i,j,1:n(3));
    end
end

k=size(X,2);

[Xall,rall]=qr([X Xhat],0);   % No pivoting here
if size(Xall,2)>n(modeX)      % Check if Xall has full modeX dimension
    Xall=Xall(:,1:n(modeX));
end
Xnew=Xall(:,k+1:end);
