
function [Iu,Iw,UIu,WIw,U,W,OutConv,F]=BlockKrylovSchurSym(A,options)
%%%%%%%%%%%%%%%%%%%%%% Krylov-Schur Code for (1,2)-Symmetric Tensor %%%%%%%%%%%%%%%%%%%%%
% function [Iu,Iw,UIu,WIw,U,W,OutConv,F]=BlockKrylovSchurSym(A,options)
%
% A   Sparse 3-tensor that is (1,2)-symmetric
%
% Method: BKS: Block Krylov-Schur for computing the best rank-(r) approximation 
% Options  r:     rank-(r_1,r_1,r_3)
%          stage: number of stages   
%          p:     block size
% Outdata 
%         Iu, Iw  : Reorderings in modes 1,2 and 3
%         UIu, WIw: Reordered solution 
%         U, W    : Solution, (not reordered)
%         OutConv : Objective function as a function of iterations
%         F       : Core tensor of dimension r
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Distributed under a 3-Clause BSD License 
%
% Copyright 2021 Lars Eldén, Linköping University, and Maryam Dehghan,
% Federal University of Ceará,
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%% More options %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if isfield(options,'r'),r=options.r; else r=[2 2 2]; end
if isfield(options,'p'),p=options.p; else p=4; end
if isfield(options,'stage'),stage=options.stage; else stage=2; end
if isfield(options,'initHOOI'),initHOOI=options.initHOOI; else initHOOI=5;end
           % initHOOI starting iterations with r=[1 1 1] before BKS is started
if isfield(options,'innerHOOI'),innerHOOI=options.innerHOOI;else,innerHOOI=0;end % HOOI for inner starting value           
if isfield(options,'tolout'),tolout=options.tolout;else,tolout=1e-13;end     % Outer stopping criterion
if isfield(options,'maxitNG'),maxitNG=options.maxitNG;else,maxitNG=20;end    % Inner maximum iterations
if isfield(options,'tolin'),tolin=options.tolin;else,tolin=1e-10;end         % Inner stopping criterion
if isfield(options,'maxit'), maxit=options.maxit; else maxit=30; end
if isfield(options,'BKSverbose'), BKSverbose=options.BKSverbose; else BKSverbose=1; end
if isfield(options,'NGverbose'), NGverbose=options.NGverbose; else NGverbose=1; end
if isfield(options,'Anonneg'), Anonneg=options.Anonneg; else Anonneg=1; end

%% Dimensions 
if BKSverbose
    disp('------------------------')
    n=size(A); disp(['Tensor dimensions:  ' num2str(n(1)) ' x ' num2str(n(2)) ' x ' num2str(n(3))]);
    disp(['BKS for r = ' num2str(r) '  Parameters, stage, p:  ' num2str([stage p])])
    disp('------------------------')
end

%% Krylov_Schur. First starting value from HOOI with r=[1 1 1]

% Random starting point of HOOI method
for i=2:3
    X0_HOOI{i}=orth(rand([n(i),1]));
end
X0_HOOI{1}=X0_HOOI{2};
% HOOI for r=[1 1 1]
[XX1, f1, nn1] = Sym_algAOI(A,X0_HOOI,[1 1 1]',initHOOI,tolout,initHOOI+2);  % itstep=maxitHOOI+2: 
                                                                                % Do not compute the gradient 

%% BKS method

uStart=XX1{1}(:,1);
% Use Block-Krylov to get first (k1,k_1,k_3) spaces
[Core_Tensor,U,W]=BKSymFirst(A,r,uStart,p,stage);  % [Core_Tensor,U,W]=Sym_mod_bk(A,r,uu,p,stage);
k=size(Core_Tensor); k=k';
if BKSverbose
    disp('------------------------')
    disp(['Dimensions of expanded core tensor: ' num2str(k')])
    disp('------------------------')
end
% Krylov-Schur loop
done=0; it=0;
while ~done && it < maxit
    it=it+1;
    %%%%%%%%%%%%%%%%%%%%%%% Compute  the (r,r,r)-rank approximation of core tensor by NG method %%%%%%%%%%%
    
    [C_svd,V_svd,W_svd]=hosvd_Truncate_rSym(Core_Tensor,r);  % HOSVD starting approximation
    
    if innerHOOI>0                                 % Initial inner HOOI iterations
        X0{1}=V_svd; X0{2}=V_svd; X0{3}=W_svd;
        [XX11,f11,nn11]=Sym_algAOI(Core_Tensor,X0,r',innerHOOI,tolout);
    else
        XX11{1}=V_svd;XX11{2}=V_svd;XX11{3}=W_svd;
    end
    
    [XX,ff,nn2]=algNlc(Core_Tensor,XX11,k,r,maxitNG,tolin,NGverbose); % Newton-Grassmann
    UXX1=U*XX{1};
    WXX3=W*XX{3};
    XXX{1}=UXX1; XXX{2}=UXX1;XXX{3}=WXX3;    % New approximation
    
    % Check convergence of outer iteration (K-S) for new approximation     
    % First do the expansion for next NG step
    [Core_Tensor,U,W]=BKSym(A,UXX1,WXX3,p,stage);
    
    % Compute gradient using the new Core_Tensor (Proposition 4.2 and figure 1 in E-D Krylov-Schur 2020)
    [F,normPhi(it),normGrad(it)]=norm_phi_gradientLocC(Core_Tensor,r);
    relative_gradient=normGrad(it)/normPhi(it);
    if relative_gradient<tolout
        done=1;
    end
end
OutConv=normGrad./normPhi;

U=U(:,1:r(1)); W=W(:,1:r(3));

if BKSverbose
    disp('------------------------')
    disp(['BKS, stage, p, k:  ' num2str(stage) '  ' num2str(p) '  '...
        num2str(k')])
    disp('------------------------')
end

 %% Sort and reorder
% Make sure that U(:,2) is the one with the sign change
% if max(U(:,1))*min(U(:,1))>=0
%     aaaaaa=1;     % Do nothing
% elseif max(U(:,2))*min(U(:,2))>=0,
%     U(:,1:2)=U(:,2:-1:1);   
% end
[uu,Iu]=sort(U(:,2));   % Minus?
    
if r(3)==1
    if Anonneg    % If r(3)=1 and the tensor is nonnegative
        if max(W)<0
            W=-W;        % make W positive
        end
    end
    Iw=1:size(W,1);
else
    [ww,Iw]=sort(W(:,2));
end

UIu=U(Iu,:);
WIw=W(Iw,:);
F=ttm(A,{U,U,W},'t');


 
 
  
  
 
